#!/usr/bin/env python

"""
Fix name-related operations. The code in this module operates upon nodes which
are produced when simplifying AST node trees originating from the compiler
module.

Copyright (C) 2006 Paul Boddie <paul@boddie.org.uk>

This software is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this library; see the file LICENCE.txt
If not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

--------

To use this module, first instantiate a Fixer object:

fixer = Fixer()

Then, apply the fixer to an existing Simplifier object:

simplifier = ...
fixer.process(simplifier)

If an existing simplifier has been used to process a module containing built-in
classes and functions, apply the fixer as follows:

fixer.process(simplifier, builtins_simplifier)
"""

from simplified import *

# Fixing of name-related operations.

class Fixer(Visitor):

    """
    The name fixer which traverses the program nodes, typically depth-first,
    and maintains a record of name usage in the different namespaces. As a
    consequence of various observations, some parts of the program node tree are
    modified with different operations employed to those originally defined.
    """

    def __init__(self):

        "Initialise the name fixer."

        Visitor.__init__(self)

        # Satisfy visitor issues.

        self.visitor = self

    def process(self, visitor, builtins_visitor=None):

        """
        Process the resources of the given 'visitor' optionally using a
        'builtins_visitor' to reference built-in objects.
        """

        self.subprograms = []
        self.current_subprograms = []
        self.current_namespaces = []

        # First, process the top-level code, finding out which names are
        # defined at that level.

        self.global_namespace = None
        self.module = visitor.result
        if builtins_visitor is not None:
            self.builtins_module = builtins_visitor.result
        else:
            self.builtins_module = None
        self.process_node(visitor.result)

        # Then, process all functions and methods, providing a global namespace.

        self.global_namespace = self.namespace

        for subprogram in visitor.subprograms:

            # Internal subprograms are skipped here and processed specially via
            # Invoke nodes.

            if not getattr(subprogram, "acquire_locals", 0):
                self.subprograms.append(self.process_node(subprogram))

        visitor.subprograms = self.subprograms
        return visitor

    def process_node(self, node, namespace=None):

        """
        Process a subprogram or module 'node', discovering from attributes on
        'node' any initial locals. Return a modified subprogram or module.
        """

        # Do not process subprograms already being processed.

        if node in self.current_subprograms:
            return None

        # Obtain a namespace either based on locals or on a structure.

        structure = structure=getattr(node, "structure", None)
        self.namespace = NameOrganiser(structure)

        # Record the current subprogram and namespace.

        self.current_subprograms.append(node)
        self.current_namespaces.append(self.namespace)

        # If passed some namespace, merge its contents into this namespace.

        if namespace is not None:
            self.namespace.merge_namespace(namespace)

        # NOTE: Check this.

        if hasattr(node, "params"):
            for param, default in node.params:
                self.namespace.store(param)
        if getattr(node, "star", None):
            param, default = node.star
            self.namespace.store(param)
        if getattr(node, "dstar", None):
            param, default = node.dstar
            self.namespace.store(param)

        # Add namespace details to any structure involved.

        if hasattr(node, "structure") and node.structure is not None:

            # Initialise bases where appropriate.

            if hasattr(node.structure, "bases"):
                bases = []
                for base in node.structure.bases:
                    bases.append(self.dispatch(base))
                node.structure.bases = bases

        # Dispatch to the code itself.

        result = self.dispatch(node)

        # Restore the previous subprogram and namespace.

        self.current_namespaces.pop()
        if self.current_namespaces:
            self.namespace = self.current_namespaces[-1]
        self.current_subprograms.pop()

        return result

    # Visitor methods.

    def default(self, node):

        """
        Process the given 'node', given that it does not have a specific
        handler.
        """

        for attr in ("args",):
            value = getattr(node, attr, None)
            if value is not None:
                setattr(node, attr, self.dispatches(value))
        for attr in ("expr", "lvalue", "test", "star", "dstar"):
            value = getattr(node, attr, None)
            if value is not None:
                setattr(node, attr, self.dispatch(value))
        for attr in ("body", "else_", "handler", "finally_", "code", "choices"):
            value = getattr(node, attr, None)
            if value is not None:
                setattr(node, attr, self.dispatches(value))
        return node

    def dispatch(self, node, *args):
        return Visitor.dispatch(self, node, *args)

    def visitGlobal(self, global_):
        for name in global_.names:
            self.namespace.make_global(name)
        return global_

    def visitLoadName(self, loadname):

        "Transform the 'loadname' node to a specific, scope-sensitive node."

        scope = self.namespace.find(loadname.name)

        # For structure namespaces, load an attribute.

        if scope == "structure":
            result = self.dispatch(LoadAttr(expr=LoadRef(ref=self.namespace.structure), name=loadname.name))

        # For global accesses (ie. those outside the local namespace)...

        elif scope == "global":

            # Where a distinct global namespace exists, examine it.

            if self.global_namespace is not None:
                scope = self.global_namespace.find(loadname.name)

                # Where the name is outside the global namespace, it must be a
                # built-in.

                if scope == "global":
                    result = self.dispatch(LoadAttr(expr=LoadRef(ref=self.builtins_module), name=loadname.name))

                # Otherwise, it is within the global namespace and must be a
                # global.

                else:
                    result = self.dispatch(LoadAttr(expr=LoadRef(ref=self.module), name=loadname.name))

            # Where no global namespace exists, we are at the module level and
            # must be accessing a built-in.

            else:
                result = self.dispatch(LoadAttr(expr=LoadRef(ref=self.builtins_module), name=loadname.name))

        # For local accesses...

        else:

            # Where a distinct global namespace exists, it must be a local.

            if self.global_namespace is not None:
                result = loadname

            # Otherwise, we must be accessing a global (which is local at the
            # module level).

            else:
                result = self.dispatch(LoadAttr(expr=LoadRef(ref=self.module), name=loadname.name))

        return result

    def visitStoreName(self, storename):

        "Transform the 'storename' node to a specific, scope-sensitive node."

        scope = self.namespace.find(storename.name)

        # For structure namespaces, store an attribute.

        if scope == "structure":
            return self.dispatch(StoreAttr(lvalue=LoadRef(ref=self.namespace.structure), name=storename.name, expr=storename.expr))

        # Where the name is outside the local namespace, disallow any built-in
        # assignment and store the name globally.

        elif scope == "global":
            return self.dispatch(StoreAttr(lvalue=LoadRef(ref=self.module), name=storename.name, expr=storename.expr))

        # For local namespace accesses...

        else:
            self.namespace.store(storename.name)

            # If a distinct global namespace exists, it must be a local access.

            if self.global_namespace is not None:
                return storename

            # Otherwise, the name is being set at the module level and is
            # considered global.

            else:
                return self.dispatch(StoreAttr(lvalue=LoadRef(ref=self.module), name=storename.name, expr=storename.expr))

    def visitInvoke(self, invoke):

        "Transform the 'invoke' node, performing processing on subprograms."

        # The special case of internal subprogram invocation is addressed by
        # propagating namespace information to the subprogram and processing it.

        if getattr(invoke, "same_frame", 0):
            subprogram = self.process_node(invoke.expr.ref, self.namespace)
            if subprogram is not None:
                self.subprograms.append(subprogram)
            return invoke
        else:
            return self.default(invoke)

class ScopeMismatch(Exception):
    pass

class NameOrganiser:

    """
    A local namespace which may either relate to a genuine set of function
    locals or the initialisation of a structure.
    """

    def __init__(self, structure=None):

        "Initialise the namespace with an optional 'structure'."

        self.structure = structure
        if structure is not None:
            self.local = "structure"
        else:
            self.local = "local"

        # Names may be self.local or "global".

        self.names = {}

    def make_global(self, name):
        if not self.names.has_key(name):
            self.names[name] = "global"
        elif self.names[name] == self.local:
            raise ScopeMismatch, "Name '%s' already considered as %s." % (name, self.local)

    def find(self, name):
        return self.names.get(name, "global")

    def store(self, name):
        if self.names.get(name) != "global":
            self.names[name] = self.local
        else:
            raise ScopeMismatch, "Name '%s' already considered as global." % name

    def merge(self, name, scope):
        if self.names.get(name) in (None, scope):
            self.names[name] = scope
        else:
            raise ScopeMismatch, "Name '%s' already considered as %s." % (name, self.names[name])

    def merge_namespace(self, namespace):
        self.merge_items(namespace.names.items())

    def merge_items(self, items):
        for name, scope in items:
            self.merge(name, scope)

    def __repr__(self):
        return repr(self.names)

# vim: tabstop=4 expandtab shiftwidth=4
